/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.service;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;

/**
 * {@link IModelQualityProvider} to count the ratio of filled out comments of
 * INamedCommentedElements.
 * 
 * @author blaschke
 * @author groh
 */
public class HierarchicElementSizeProvider implements IModelQualityProvider<IHierarchicElement> {
	/**
	 * returns an array of integers or <String,Double> Map Number of children of the
	 * hierarchical Element (on this level) Number of ports on the hierarchical
	 * Element (on this level) Number of channels in the hierarchical element (on
	 * this level) Number of totalPorts in whole tree Number of totalElements in
	 * whole tree Average SubTree Size
	 */

	/** Key for the metric counting ports */
	public static final String NUMBER_OF_PORTS = "numberOfPorts";

	/** Key for the metric counting elements */
	public static final String NUMBER_OF_CONTAINED_ELEMENTS = "numberOfContainedElements";

	/** Key for the metric counting channels */
	public static final String NUMBER_OF_CHANNELS = "numberOfChannels";

	/** Key for the metric counting the total amount of ports contained */
	public static final String NUMBER_OF_TOTAL_PORTS = "numberOfTotalPorts";
	/** Key for the metric counting the total amount of elements contained */
	public static final String NUMBER_OF_TOTAL_ELEMENTS = "numberOfTotalElements";
	/** Key for the metric counting the total amount of elements which can be commented */
	public static final String NUMBER_OF_TOTAL_COMMENTABLE_ELEMENTS =
			"numberOfTotalCommentableElements";
	/** Key for the metric counting the total amount of elements which are commented */
	public static final String NUMBER_OF_TOTAL_COMMENTED_ELEMENTS =
			"numberOfTotalCommentedElements";

	/** Map containing all metrics concerning the current IHierarchicElement */
	private Map<String, Double> storedMetrics;
	/**
	 * List containing all HierarchicElementSizeProvider for components inside the
	 * IHierarchicElement
	 */
	private List<HierarchicElementSizeProvider> containedProviders;

	/**
	 * Mapping containing all providers from this element and any elements contained in that element
	 */
	private Map<IHierarchicElement, HierarchicElementSizeProvider> allProviders;

	/**
	 * @return Map containing the mapping between a metric keys and the corresponding values for
	 *         this element
	 */
	public Map<String, Double> getStoredMetrics() {
		return storedMetrics;
	}

	/** */
	public HierarchicElementSizeProvider() {
		super();
		this.storedMetrics = new HashMap<>();
		this.allProviders = new HashMap<>();
		this.containedProviders = new ArrayList<>();
	}

	/** {@inheritDoc} */
	@Override
	public void apply(IHierarchicElement currentElement) {

		storedMetrics.put(NUMBER_OF_PORTS, (double)currentElement.getConnectors().size());
		storedMetrics.put(NUMBER_OF_CONTAINED_ELEMENTS,
				(double)currentElement.getContainedElements().size());
		storedMetrics.put(NUMBER_OF_CHANNELS, (double)currentElement.getConnections().size());

		// depth metrics
		storedMetrics.put(NUMBER_OF_TOTAL_PORTS, (double)currentElement.getConnectors().size());
		storedMetrics.put(NUMBER_OF_TOTAL_ELEMENTS, 1.0);

		boolean isElementCommentable = currentElement instanceof INamedCommentedElement;
		storedMetrics.put(NUMBER_OF_TOTAL_COMMENTABLE_ELEMENTS, isElementCommentable ? 1.0 : 0.0);
		String comment =
				isElementCommentable ? ((INamedCommentedElement)currentElement).getComment() : null;
		storedMetrics.put(NUMBER_OF_TOTAL_COMMENTED_ELEMENTS,
				comment != null && comment != "" ? 1.0 : 0.0);

		for(IHierarchicElement containedElement : currentElement.getContainedElements()) {
			HierarchicElementSizeProvider provider = new HierarchicElementSizeProvider();
			containedProviders.add(provider);
			provider.apply(containedElement);
			allProviders.putAll(provider.allProviders);

			for(String key : new String[] {NUMBER_OF_TOTAL_PORTS, NUMBER_OF_TOTAL_ELEMENTS,
					NUMBER_OF_TOTAL_COMMENTABLE_ELEMENTS, NUMBER_OF_TOTAL_COMMENTED_ELEMENTS}) {
				storedMetrics.merge(key, provider.storedMetrics.get(key), Double::sum);
			}
		}
		allProviders.put(currentElement, this);
	}

	/** {@inheritDoc} */
	@Override
	public String getExampleString() {
		double commentable = storedMetrics.getOrDefault(
				HierarchicElementSizeProvider.NUMBER_OF_TOTAL_COMMENTABLE_ELEMENTS, 0.0);
		double commented = storedMetrics.getOrDefault(
				HierarchicElementSizeProvider.NUMBER_OF_TOTAL_COMMENTED_ELEMENTS, 0.0);

		// Create a DecimalFormat object to format the percentage
		DecimalFormat decimalFormat = new DecimalFormat("0.00");

		// Format the percentage value and append the '%' symbol
		String formattedPercentage = decimalFormat.format((commented / commentable) * 100) + "%";

		return "Commented elements: " + (int)commented + ", Total elements: " + (int)commentable +
				", Ratio: " + formattedPercentage;
	}

	/** {@inheritDoc} */
	@Override
	public IModelQualityProvider<IHierarchicElement> getProvider(IHierarchicElement element) {
		return allProviders.get(element);
	}

	/** {@inheritDoc} */
	@Override
	public void traverse(IVisitorModelQualityProvider visitor) {
		
		
		// iterate through all contained providers
		for(HierarchicElementSizeProvider provider : containedProviders) {
			provider.traverse(visitor);
		}
		
		// visit this provider
		visitor.accept(this);
	}

	/** {@inheritDoc} */
	@Override
	public List<String> metricKeys() {
		return new ArrayList<>(storedMetrics.keySet());
	}

	/*
	 * 
	 * CLONE DETECTION START public class CloneDetection {
	 * 
	 * private Map<Integer, List<IHierarchicElement>> map = new HashMap<>();
	 * 
	 * public void printResult() { System.out.println("Stuff"); }
	 * 
	 * public void apply(IHierarchicElement element) { int hash = 1; for(IConnector
	 * connector : element.getConnectors()) { EList<IModelElementSpecification>
	 * specifications = connector.getSpecifications(); if(!specifications.isEmpty()
	 * && connector.getSpecifications().get(0) instanceof PortSpecification) {
	 * 
	 * PortSpecification port =
	 * (PortSpecification)connector.getSpecifications().get(0); hash = 31 * hash +
	 * (connector instanceof InputPort ? 1 : 2) *
	 * port.getType().getClass().hashCode(); } else { // Clone Detection for State
	 * diagrams not supported return; } }map.putIfAbsent(hash,new ArrayList<>());
	 * 
	 * List<IHierarchicElement> elements = map.get(hash);
	 * 
	 * for( IHierarchicElement element_b:elements) {
	 * 
	 * List<String> element_a_input = new ArrayList<String>(); List<String>
	 * element_b_input = new ArrayList<String>(); List<String> element_a_output =
	 * new ArrayList<String>(); List<String> element_b_output = new
	 * ArrayList<String>();
	 * 
	 * for(IConnector connector : element.getConnectors()) { PortSpecification port
	 * = (PortSpecification)connector.getSpecifications().get(0); String type =
	 * port.getType().getClass().getSimpleName(); if(connector instanceof InputPort)
	 * { element_a_input.add(type); } else { element_a_output.add(type); } }
	 * 
	 * for(IConnector connector : element_b.getConnectors()) { PortSpecification
	 * port = (PortSpecification)connector.getSpecifications().get(0); String type =
	 * port.getType().getClass().getSimpleName(); if(connector instanceof InputPort)
	 * { element_b_input.add(type); } else { element_b_output.add(type); } }
	 * 
	 * boolean equals = true;
	 * 
	 * for(String type_a_input : element_a_input) {
	 * if(!element_b_input.remove(type_a_input)) { equals = false; break; }
	 * 
	 * }
	 * 
	 * for(String type_a_output : element_a_output) {
	 * if(!element_b_output.remove(type_a_output)) { equals = false; break; }
	 * 
	 * } if(equals) { System.out .println("Found duplicate: " +
	 * ((INamedCommentedElement)element).getName() + ", " +
	 * ((INamedCommentedElement)element_b).getName()); break; } else {
	 * System.out.println("Same but not duplicate"); } }elements.add(element); }
	 * CLONE DETECTION END
	 * 
	 * 
	 * public void printShare(HierarchicElementBase element) {
	 * 
	 * Set<Tuple<String, Integer>> counts = new HashSet<>(); this.apply(element);
	 * counts.add(new Tuple<>("[Self]" + element.getName(), count)); this.count = 0;
	 * 
	 * for(IHierarchicElement child : element.getContainedElements()) {
	 * 
	 * HierarchicElementBase real_child = (HierarchicElementBase)child;
	 * 
	 * recursiveDescent(real_child, this);
	 * 
	 * counts.add(new Tuple<>(real_child.getName(), count)); count = 0; }
	 * 
	 * double total = counts.stream().mapToInt(t -> t.getElement2()).sum();
	 * 
	 * System.out.println("[" + this.getClass().getSimpleName() + "] shares:");
	 * for(Tuple<String, Integer> tuple : counts) { System.out.println("\t " +
	 * String.format("%05.2f%%", tuple.element2 / total * 100) + ", " +
	 * tuple.element2 + ": " + tuple.element1); } }
	 * 
	 */
}
